package edu.usc.middleware.shibboleth.idp;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.httpclient.HttpStatus;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.chrono.ISOChronology;
import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.ISODateTimeFormat;
import org.opensaml.xml.security.x509.X509Credential;
import org.opensaml.xml.util.Base64;
import org.opensaml.xml.util.DatatypeHelper;
import org.opensaml.xml.util.LazyList;
import org.opensaml.util.storage.StorageService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.net.httpserver.HttpContext;

import edu.internet2.middleware.shibboleth.common.attribute.resolver.AttributeResolutionException;
import edu.internet2.middleware.shibboleth.common.attribute.resolver.AttributeResolver;
import edu.internet2.middleware.shibboleth.common.relyingparty.RelyingPartyConfiguration;
import edu.internet2.middleware.shibboleth.common.relyingparty.RelyingPartyConfigurationManager;
import edu.internet2.middleware.shibboleth.idp.util.HttpServletHelper;
import edu.internet2.middleware.shibboleth.idp.util.IPRange;

import org.terracotta.api.ClusteringToolkit;
import org.terracotta.api.Terracotta;
import org.terracotta.cluster.*;

/***/
/**
 * A Servlet for verifying the aliveness of the IdP. 
 * 
 * This class uses many code examples from
 * edu.internet2.middleware.shibboleth.idp.StatusServlet except that only the
 * word AVAILABLE is printed if all of the checks succeed.
 * 
 * If the checks do not succeed, then HTTP code 503 (Service Unavailable) is
 * thrown with a message about which underlying service was not responding
 * correctly.
 * 
 * The success page was designed to be a copy of the html which would be generated
 * from the original status jsp which Nagios and the loadbalancer are already
 * programmed to work with.
 */
public class StatusServlet extends HttpServlet {

    /** Serial version UID. */
//    private static final long serialVersionUID = -5280549109235107879L;

    private final String IP_PARAM_NAME = "AllowedIPs";
    
    private final Logger log = LoggerFactory.getLogger(StatusServlet.class);

    private LazyList<IPRange> allowedIPs;

    /** Formatter used when print date/times. */
    private DateTimeFormatter dateFormat;

    /** Time the IdP started up. */
    private DateTime startTime;

    /** Attribute resolver used by the IdP. */
    private AttributeResolver<?> attributeResolver;

    /** Relying party configuration manager used by the IdP. */
    private RelyingPartyConfigurationManager rpConfigManager;

    /** Storage service used by the IdP. */
    private StorageService<?,?> store;
    
	/**
	 * Flag for clustered operations availablility. Defaults to true unless the
	 * cluster listener indicates the cluster is down. Always is true when the
	 * IdP was started without the use of clustering (and hence clustered
	 * operations issues won't be a cause for a status alarm).
	 */
    private boolean clusteredOperationsEnabled = true;

    /** {@inheritDoc} */
    public void init(ServletConfig config) throws ServletException {
        super.init(config);

        allowedIPs = new LazyList<IPRange>();

        String cidrBlocks = DatatypeHelper.safeTrimOrNullString(config.getInitParameter(IP_PARAM_NAME));
        if (cidrBlocks != null) {
            for (String cidrBlock : cidrBlocks.split(" ")) {
                allowedIPs.add(IPRange.parseCIDRBlock(cidrBlock));
            }
        }

        dateFormat = ISODateTimeFormat.dateTimeNoMillis();
        startTime = new DateTime(ISOChronology.getInstanceUTC());
        attributeResolver = HttpServletHelper.getAttributeResolver(config.getServletContext());
        rpConfigManager = HttpServletHelper.getRelyingPartyConfirmationManager(config.getServletContext());
        store = HttpServletHelper.getStorageService(config.getServletContext());

        if (Terracotta.isActive()) {
            log.debug("Terracotta is active.  Setting up cluster listener for operational availability.");
            ClusterInfo clusterInfo = new TerracottaClusterInfo();
            final StatusServlet thisServ = this;
            clusterInfo.addClusterListener(new ClusterListener()
              {
                 public void nodeJoined(ClusterEvent event) {
//                     log.debug("Node Joined");
                 }
                 public void nodeLeft(ClusterEvent event) {
//                     log.debug("Node Left");
                 }
                 public void operationsEnabled(ClusterEvent event) {
                     log.info("Cluster operations enabled");
                     setClusteredOperationsEnabled(true);
                 }
                 public void operationsDisabled(ClusterEvent event) {
                     log.error("Cluster operations disabled!");
                     setClusteredOperationsEnabled(false);
                 }
            });
        }
        else
            log.debug("Terracotta clustering is not enabled.  No cluster listener set.");
        
    }

    /** {@inheritDoc} */
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        if (!isAuthenticated(request)) {
            response.sendError(HttpStatus.SC_UNAUTHORIZED);
            return;
        }

        response.setContentType("text/html");
        PrintWriter output = response.getWriter();

        if (IdPAlivenessCheck()) {
            // set status header
            response.setHeader("X-Shibboleth-Status", "Available");

            HttpSession session = request.getSession();
            // if they didn't already have a session, let's not create one
            if (session.isNew()) session.invalidate(); 

			// This html code mimics the original status jsp page which Nagios
			// and the loadbalancer are already set to process
            output.println("<html>");
            output.println("  <head>");
            output.println("    <title>Available</title>");
            output.println("  </head>");
            output.println("  <body>");
            output.println("");
            output.println("AVAILABLE");
            output.println("<!-- be533d557c15a164a1c8b64545231abd -->");
            output.println("");
            output.println("");
            output.println("  </body>");
            output.println("</html>");
        	output.flush();
        	log.debug("Status check from IP '{}' succeeded", request.getRemoteAddr());
        }
        else {
            log.error("Status check from IP '{}' failed.  IDP IS DOWN!", request.getRemoteAddr());
        	response.sendError(HttpStatus.SC_SERVICE_UNAVAILABLE, clusteredOperationsEnabled ? "Attribute resolver failure -- LDAP issue" : "Cluster inoperational.");
        	return;
        }
        
        output.flush();
    }
    
    

    /**
     * Checks whether the client is authenticated.
     * 
     * @param request client request
     * 
     * @return true if the client is authenticated, false if not
     */
    protected boolean isAuthenticated(HttpServletRequest request) throws ServletException {
        log.debug("Attempting to authenticate client '{}'", request.getRemoteAddr());
        try {
            InetAddress clientAddress = InetAddress.getByName(request.getRemoteAddr());

            for (IPRange range : allowedIPs) {
                if (range.contains(clientAddress)) {
                    return true;
                }
            }

            return false;
        } catch (UnknownHostException e) {
            throw new ServletException(e);
        }
    }

    

    /**
     * Performs a number of checks to determine if the IdP is actually alive. If any of the checks fail but do not hang,
     * false will be returned. If all checks succeed, true will be returned. The check may hang when attempting to
     * access the storage service, but this is ok because lack of a response will also alert Nagios and the loadbalancer
     * that the service is down.
     */
    protected boolean IdPAlivenessCheck() {
        try {
            attributeResolver.validate();
            log.debug("Attribute resolver validated.  Attribute store operational.");
//            out.println("attribute_resolver_valid: " + Boolean.TRUE);
        } catch (AttributeResolutionException e) {
//            out.println("attribute_resolver_valid: " + Boolean.FALSE);
            log.debug("Attribute resolver NOT validated.  Attribute store INOPERATIONAL!");
            return false;
        }
        if (Terracotta.isActive()) {
			if (clusteredOperationsEnabled)
				log.debug("Cluster is functional.");
			else
				log.debug("Cluster is MALFUNCTIONING!");
	        return clusteredOperationsEnabled;
        }
        return true;
    }

    
    synchronized protected void setClusteredOperationsEnabled(boolean val) {
        clusteredOperationsEnabled = val;
    }
}